#
#  This file is part of the BayesicFitting package.
#
#  BayesicFitting is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of
#  the License, or (at your option) any later version.
#
#  BayesicFitting is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#


from .Biweight import Biweight
from .CosSquare import CosSquare
from .Cosine import Cosine
from .Gauss import Gauss
from .Huber import Huber
from .Kernel import Kernel
from .Lorentz import Lorentz
from .Parabola import Parabola
from .Sinc import Sinc
from .Triangle import Triangle
from .Tricube import Tricube
from .Triweight import Triweight
from .Uniform import Uniform

