#!/usr/bin/env python
# -*- coding: utf-8 -*-
# PYTHON.test -> test area for basic fitter package

"""
from Test2dFixedModels import Test2dFixedModels
from Test2dModels import Test2dModels
# from TestAll import TestAll
from TestAmoeba import TestAmoeba
from TestAmoebaFitter import TestAmoebaFitter
from TestBaseFitter import TestBaseFitter
from TestBaseModel import TestBaseModel
from TestBracketModel import TestBracketModel
from TestCombiModel import TestCombiModel
from TestCompoundModel import TestCompoundModel
from TestCurveFitter import TestCurveFitter
from TestDriftModels import TestDriftModels
from TestEngine import TestEngine
from TestEngine2 import TestEngine2
from TestEngine3 import TestEngine3
from TestErrorDistribution import TestErrorDistribution
from TestEvidence import TestEvidence
from TestEvidence2 import TestEvidence2
from TestFitter import TestFitter
from TestFixedModels import TestFixedModels
from TestFormatter import TestFormatter
from TestFreeShape2dModel import TestFreeShape2dModel
from TestFreeShapeModel import TestFreeShapeModel
from TestImageAssistant import TestImageAssistant
from TestKernel2dModel import TestKernel2dModel
from TestKernels import TestKernels
from TestLMFitter import TestLMFitter
from TestLogFactorial import TestLogFactorial
from TestModels import TestModels
from TestMonteCarlo import TestMonteCarlo
from TestNLFitters import TestNLFitters
from TestNestedSampler import TestNestedSampler
from TestPrior import TestPrior
from TestQRFitter import TestQRFitter
from TestRobustShell import TestRobustShell
from TestSampleList import TestSampleList
from TestStartEngine import TestStartEngine
from TestTools import TestTools
from TestWeights import TestWeights

print( "__init__  done" )
"""
